# core-image-tester
# Copyright (C) 2015 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import enum
import logging
from pdb import bdb

import kombu
from kombu.mixins import ConsumerMixin

from core_image_tester import constants

logger = logging.getLogger(__name__)

__all__ = [
    'MessageActions',
    'CoreImageTesterQueueMonitor',
    'create_connection_from_config',
    'CoreImageTestResultPublisher',
]


class MessageActions(enum.Enum):

    """Actions that the message processing CB can request.

    These enum values can be returned from the message processing callback,
    and the CoreImagePublisherQueueMonitor will take appropriate action.

    """

    # The message needs to be retried. This might cause the message to be
    # retried, or it might reject the message if it's retry count has been
    # exceeded.
    Retry = 1

    # The message was processed properly, and should be acknowledged.
    Acknowledge = 2


class CoreImageTesterQueueMonitor(ConsumerMixin):

    """A class that watches for incoming messages from a queue, and calls
    a callback to process them.

    By using an external callback we can separate the stateful logic of
    connecting to amqp from our business logic of dealing with the messages.

    """

    def __init__(self, connection, on_message_cb):
        """Construct a new CoreImagePublisherQueueMonitor.

        :param connection: A kombu.Connection instance.
        :param on_message_db: A callable that will be called for every new
            messages from the rabbitMQ queue(s). It must accept a single
            parameter: the message object. The callback MUST either acknowledge
            or reject or requeue the messsage, and MUST NOT raise any unhandled
            exceptions.

        """
        self.connection = connection
        self.on_message_cb = on_message_cb

    def get_consumers(self, Consumer, channel):
        """Return consumers instances for all configured queues."""
        queues = [
            kombu.Queue('core.tests.{}'.format(constants.API_VERSION))
        ]
        return [Consumer(queues=queues, callbacks=[self.process])]

    def process(self, body, message):
        """Process incomming test request.

        Run requested tests and posts results to the 'adt_results' queue
        for later checking.
        """
        extra = constants.LOGGING_EXTRA.copy()
        extra.update(body)
        logger.info('Got: %r', body, extra=extra)
        try:
            requested_action = self.on_message_cb(body)
            if requested_action == MessageActions.Retry:
                self.maybe_requeue_message(message)
            elif requested_action == MessageActions.Acknowledge:
                message.ack()
        # Make it possible to use a debugger inside the worker callback:
        except bdb.BdbQuit:
            raise
        except Exception as e:
            logger.error(
                "Caught unhandled exception while processing message: %s",
                e,
                extra=extra
            )
            message.requeue()

    def maybe_requeue_message(self, message):
        """Maybe requeue the message, based on it's retry count.

        Note: kombu (or maybe rabbitmq) does not allow you to alter a message's
        payload and then call requeue(), which would be super helpful here.
        Instead we explicitly queue it (using kombu.SimpleQueue) to the last
        queue the message was on.

        The key we use to store the retry count is based on the message's
        routing key, which allows us to have a separate retry count for each
        stage of the system.

        """
        extra = constants.LOGGING_EXTRA.copy()
        extra.update(message.payload)
        retry_key = "{}_retry_count".format(
            message.delivery_info.get('routing_key', '')
        )
        retry_count = int(message.payload.get(retry_key, '0'))
        if retry_count < constants.RETRY_COUNT:
            message.payload[retry_key] = retry_count + 1
            queue = self.connection.SimpleQueue(
                'core.tests.{}'.format(constants.API_VERSION))
            queue.put(message.payload)
            queue.close()
            message.ack()
        else:
            logger.error(
                "Rejecting message due to retry count exceeding maximum. "
                "Message will reside on the dead letter queue",
                extra=extra
            )
            queue = self.connection.SimpleQueue(
                'core.deadletters.{}'.format(constants.API_VERSION))
            queue.put(message.payload)
            queue.close()
            message.ack()


def create_connection_from_config(config):
    """Create a connection given a config object.

    'config' can either be a ConfigParser instance, or a nested dictionary.

    """
    amqp_uris = config.get('amqp', 'uris').split()
    return kombu.Connection(amqp_uris)


class CoreImageTestResultPublisher(object):

    """A callable that knows how to publish results from this service.

    For now we're using a simple queue, but this can easily be extended to use
    a full-blown topic exchange in the future.

    """

    def __init__(self, connection):
        self.connection = connection

    def __call__(self, payload):
        """Take 'payload' and enqueue it on the rabbit queue."""
        queue = self.connection.SimpleQueue(
            "core.result.{}".format(constants.API_VERSION)
        )
        queue.put(payload)
        queue.close()
