# core-image-tester
# Copyright (C) 2015 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"Utilities that might be useful elsewhere."

import subprocess
import textwrap


def check_call(*popenargs, **kwargs):
    """A better version of subprocess.check_call that gives better errors.

    If a command invoked via 'check_call' fails, the error looks like::

        Command '['juju', 'bootstrap']' returned non-zero exit status 2

    While concise, that doesn't help the user of this script see what went
    wrong. This function invokes a command, but captures stdout and stderr,
    and includes that information in the exception.

    """
    if 'stdout' in kwargs or 'stderr' in kwargs:
        # user did something with stdout or stderr already, so just call
        # through to check_call directly:
        return subprocess.check_call(*popenargs, **kwargs)
    kwargs['stdout'] = kwargs['stderr'] = subprocess.PIPE
    process = subprocess.Popen(*popenargs, **kwargs)
    stdout, stderr = process.communicate()
    if process.returncode != 0:
        raise BetterCalledProcessError(
            process.returncode,
            popenargs[0],
            textwrap.dedent(
                '''
                stdout:
                """\\
                {}"""
                stderr:
                """\\
                {}"""
                ''').format(stdout.decode(), stderr.decode())
        )


class BetterCalledProcessError(subprocess.CalledProcessError):

    """Extend CalledProcessError to show failed command output as well."""

    def __str__(self):
        return super().__str__() + self.output
